/* This file is a component of the JunkBusterMT Proxy, based on original code from
 * the Internet Junkbuster Proxy.
 *
 * Copyright (C) 2000 Gabriel L. Somlo
 * Copyright (C) 1997 Anonymous Coders and Junkbusters Corporation
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see ../GPL.txt); if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/types.h>

#ifndef WIN_GCC
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
#else
#include <io.h>
#include <windows.h>
#endif

#include "parsers.h"
#include "filegard.h"


#ifdef WIN_GCC

int
winsock_init( void ) {

  WORD wVersionRequested;
  WSADATA wsaData;

  wVersionRequested = MAKEWORD( 2, 0 );
  return( WSAStartup(wVersionRequested, &wsaData) );

} /* winsock_init */

/* fake gethostbyname_r function for windows. May need mutex locks(GLS) */
static int
gethostbyname_r_win32( char *h, struct hostent *he, int *e ) {

  struct hostent *hp;

  if( !(hp = gethostbyname(h)) ) {
    *e = EINVAL;
    return( -EINVAL );
  }
  memcpy( he, hp, sizeof(struct hostent) );
  *e = 0;
  return( 0 );

} /* gethostbyname_r_win32 */

#define EPROTOTYPE WSAEPROTOTYPE	/* fake EPROTOTYPE errno for windows */
#define EADDRINUSE WSAEADDRINUSE	/* fake EADDRINUSE errno for windows */
#define EINPROGRESS WSAEINPROGRESS	/* fake EINPROGRESS errno for windows */

#endif /* WIN_GCC */


int
write_socket( int fd, char *buf, int n ) {
#ifdef WIN_GCC
  return( send(fd, buf, n, 0) );
#else
  return( write(fd, buf, n) );
#endif
} /* write_socket */


int
read_socket( int fd, char *buf, int n ) {
#ifdef WIN_GCC
  return( recv(fd, buf, n, 0) );
#else
  return( read(fd, buf, n) );
#endif
} /* read_socket */


void
close_socket( int fd ) {
#ifdef WIN_GCC
  closesocket( fd );
#else
  close( fd );
#endif
} /* close_socket */


int
atoip( char *host ) {

  struct sockaddr_in inaddr;
  struct hostent host_ent;
#ifndef WIN_GCC
  char buf[BUFSIZ];
#endif
  int h_errnop;
#ifdef LIN_GCC
  struct hostent *host_p;
#endif

  if( !host || (*host == '\0') ) 
    return( INADDR_ANY );

  memset( &inaddr, 0, sizeof(inaddr) );

  if( (inaddr.sin_addr.s_addr = inet_addr(host)) == -1 ) {

#if defined( LIN_GCC )
    if( gethostbyname_r(host, &host_ent, buf, BUFSIZ, &host_p, &h_errnop) != 0 )
#elif defined( SOL_GCC )
    if( gethostbyname_r(host, &host_ent, buf, BUFSIZ, &h_errnop) == NULL )
#elif defined( WIN_GCC )
    if( gethostbyname_r_win32(host, &host_ent, &h_errnop) != 0 )
#else
  #error "gethostbyname_r(): Only Linux, Solaris, and Win32 handled for now..."
#endif
    {
      errno = h_errnop;
      return( -1 );
    }

    if( host_ent.h_addrtype != AF_INET ) {
      errno = EPROTOTYPE;
      return( -1 );
    }
    memcpy( &inaddr.sin_addr, host_ent.h_addr, sizeof(inaddr.sin_addr) );
  }

  return( inaddr.sin_addr.s_addr );

} /* atoip */




/* start bind.c */

/*
 * BIND-PORT (portnum)
 *  if success, return file descriptor
 *  if failure, returns -2 if address is in use, otherwise -1
 */
int
bind_port( char *hostnam, int portnum ) {

  int sd;
  struct sockaddr_in sa;
  int one = 1;

  memset( &sa, '\0', sizeof(struct sockaddr_in) );

  if( (sd = socket(AF_INET, SOCK_STREAM, 0)) < 0 )
    return( -1 );

  sa.sin_family = AF_INET;
  sa.sin_addr.s_addr = atoip( hostnam );
  if( sizeof(sa.sin_port) == sizeof(short) )
    sa.sin_port = htons( portnum );
  else
    sa.sin_port = htonl( portnum );

  setsockopt( sd, SOL_SOCKET, SO_REUSEADDR, (char *)&one, sizeof(one) );

  if( bind(sd, (struct sockaddr *)&sa, sizeof(struct sockaddr_in)) < 0) {
    close_socket( sd );
    if( errno == EADDRINUSE )
      return( -2 );
    else
      return( -1 );
  }

  while( listen(sd, 5) == -1 )
    if( errno != EINTR )
      return( -1 );

  return( sd );

} /* bind_port */


/* 
 * ACCEPT-CONNECTION
 * the argument, fd, is the value returned from bind_port
 *
 * when a connection is accepted, it returns the file descriptor
 * for the connected port
 */
int
accept_connection( int sd, char **p_addr_str, long *p_addr_long ) {

  int asd;
  struct sockaddr sa;
  struct sockaddr_in *rap;
  int sa_len;

  rap = (struct sockaddr_in *)&sa;

  sa_len = sizeof( struct sockaddr );

  do {
    asd = accept( sd, &sa, &sa_len );
  } while( (asd < 1) && (errno == EINTR) );

  if( asd < 0 )
    return( -1 );

  *p_addr_str  = strdup( inet_ntoa(rap->sin_addr) );
  *p_addr_long = ntohl( rap->sin_addr.s_addr );

  return asd;

} /* accept_connection */


/* end bind.c */
/* start conn.c */


int
connect_to( char *host, int portnum, long addr_long ) {

  struct sockaddr_in inaddr;
  int fd, addr;
  fd_set wfds;
  struct timeval tv;
#ifndef WIN_GCC
  int flags;
#endif
  struct access_control_addr src[1], dst[1];
  int one = 1;

  memset( &inaddr, 0, sizeof(struct sockaddr_in) );

  if( (addr = atoip(host)) == -1)
    return( -1 );

  src->addr = addr_long;
  src->port = 0;

  dst->addr = ntohl( addr );
  dst->port = portnum;

  if( block_acl(src, dst) ) {
    errno = EPERM;
    return( -1 );
  }

  inaddr.sin_addr.s_addr = addr;
  inaddr.sin_family      = AF_INET;

  if( sizeof(inaddr.sin_port) == sizeof(short) )
    inaddr.sin_port = htons( portnum );
  else
    inaddr.sin_port = htonl( portnum );

  if( (fd = socket(inaddr.sin_family, SOCK_STREAM, 0)) < 0 )
    return( -1 );

  /* turn off TCP coalescence */
  setsockopt( fd, IPPROTO_TCP, TCP_NODELAY, (char *)&one, sizeof(int) );

#ifndef WIN_GCC
  if( (flags = fcntl(fd, F_GETFL, 0)) != -1 ) {
    flags |= O_NDELAY;
    fcntl( fd, F_SETFL, flags );
  }
#endif

  while( connect(fd, (struct sockaddr *)&inaddr, sizeof(inaddr) ) == -1 ) {
    if( errno == EINPROGRESS )
      break;
    if( errno != EINTR ) {
      close_socket( fd );
      return( -1 );
    }
  }

#ifndef WIN_GCC
  if( flags != -1 ) {
    flags &= ~O_NDELAY;
    fcntl( fd, F_SETFL, flags );
  }
#endif

  /* wait for connection to complete */
  FD_ZERO( &wfds );
  FD_SET( fd, &wfds );

  /* connection timeout interval (?) */
  tv.tv_sec  = 50;	/* was 30 */
  tv.tv_usec = 0;

  if( select(fd + 1, NULL, &wfds, NULL, &tv) <= 0 ) {
    close_socket( fd );
    return( -1 );
  }

  return( fd );

} /* connect_to */


int
direct_connect( struct gateway *gw,
                struct http_request *http,
                long addr_long ) {

  if( gw->forward_host )
    return( connect_to(gw->forward_host, gw->forward_port, addr_long) );
  else
    return( connect_to(http->host, http->port, addr_long) );

} /* direct_connect */
