/* This file is a component of the JunkBusterMT Proxy, based on original code from
 * the Internet Junkbuster Proxy.
 *
 * Copyright (C) 2000 Gabriel L. Somlo
 * Copyright (C) 1997 Anonymous Coders and Junkbusters Corporation
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see ../GPL.txt); if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/*********************************  errlog.c  **********************************

    Purpose: Log errors to a designated destination in an elegant, printf-like
             fashion.

    Notes:   none.
**/


#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <pthread.h>

#include "errlog.h"
#include "bindconn.h"


static char *progname;			/* name of the program -> argv[0] */
static FILE *logfp = NULL;		/* where to log (default: stderr) */
static int debug = ERR | ERN;		/* logging detail level */


/*---------------------------------------------------------------------*/
/* initialize the logging module;                                      */
/* Arguments: char *logfname - in - the filename to log to;            */
/*            int debuglevel - in - the intended logging detail level; */
/* Notes: If the logfname is NULL, the logfile defaults to stderr...   */
/*---------------------------------------------------------------------*/
void
init_errlog( char *prog_name, char *logfname, int debuglevel ) {

  FILE *fp;

  /* set the program name */
  progname = strdup( prog_name );

  /* set the logging detail level */
  debug |= debuglevel;			/* maintain the defaults! */

  /* set the designated log file */
  if( logfname ) {
    if( !(fp = fopen(logfname, "a")) ) {
      errlogf( ERN, "init_errlog(): can't open logfile: %s\n", logfname );
      exit( 1 );
    }
    logfp = fp;
  } else {
    logfp = stderr;
  }

  /* set logging to be completely unbuffered */
  setbuf( logfp, NULL );

} /* init_errlog */


/*--------------------------------------------------------------------------*/
/* This is the error-reporting and logging function.                        */
/* Arguments: int loglevel - in - the type of message to be logged;         */
/*            char *fmt - in - the main string we want logged, printf-like; */
/*            ... - in - arguments to be inserted in fmt (printf-like);     */
/* Notes:                (the one exception is for LOG, where fmt is a raw  */
/*                        buffer, and the first var. arg. is its length)    */
/*--------------------------------------------------------------------------*/
void
errlogf( int loglevel, char *fmt, ... ) {

  va_list ap;
  char *p, outbuf[BUFSIZ];
  int outc = 0;
  long this_thread;	/* was: pthread_t this_thread;*/

  /* deal with glibc stupidity - it won't let you initialize logfp */
  if( !logfp )
    logfp = stderr;

  /* verify if loglevel applies to current settings and bail out if negative */
  if( !(loglevel & debug) )
    return;

  /* get ready to scan var. args. */
  va_start( ap, fmt );

  /* if request is LOG, dump fmt knowing size is first var. arg. */
  if( loglevel == LOG ) {

    int ival;

    ival = va_arg( ap, int );		/* get size of buffer from first arg. */
    fwrite( fmt, ival, 1, logfp );
    va_end( ap );
    return;

  }

  /* HDR, CON, GPC, and ERR need a fully formated/cooked fmt +var. args */

  /* initialize outbuf */
  memset( outbuf, '\0', BUFSIZ );

  /* build formatted message from fmt and var-args */
  for( p = fmt; *p; ++p ) {

    char tempbuf[BUFSIZ];
    char *sval;
    int ival;
    char cval;

    if( *p != '%' ) {
      outbuf[outc++] = *p;
      continue;
    }
    switch( *++p ) {
      case 'd':
        ival = va_arg( ap, int );
        sprintf( tempbuf, "%d", ival );
        strcat( outbuf, tempbuf );
        if( (outc += strlen(tempbuf)) >=BUFSIZ ) {
          fprintf( logfp, "Surfagent: ERR: errlogf(): buffer overflow - d. "
                          "Exiting...\n\n" );
          exit( 1 );
        }
        break;
      case 'c':
        cval = va_arg( ap, char );
        sprintf( tempbuf, "%c", cval );
        strcat( outbuf, tempbuf );
        if( (outc += strlen(tempbuf)) >=BUFSIZ ) {
          fprintf( logfp, "Surfagent: ERR: errlogf(): buffer overflow - c. "
                          "Exiting...\n\n" );
          exit( 1 );
        }
        break;
      case 's':
        sval = va_arg( ap, char * );
        sprintf( tempbuf, "%s", sval );
        strcat( outbuf, tempbuf );
        if( (outc += strlen(tempbuf)) >=BUFSIZ ) {
          fprintf( logfp, "Surfagent: ERR: errlogf(): buffer overflow - s. "
                          "Exiting...\n\n" );
          exit( 1 );
        }
        break;
      default:
        fprintf( logfp, "SurfAgent: ERR: errlogf(): unknown format \'%c\'\n"
                        "Current log buffer is: %s\nExiting...\n\n",
                        *p, outbuf );
        exit( 1 );
        break;

    } /* switch( p ) */

  } /* for( p ... ) */

  /* done with var. args */
  va_end( ap );

  /* get current thread id */
  this_thread = (long)pthread_self();

  /* use the outbuf according to the command (loglevel) */
  switch( loglevel ) {
    case HDR:
      fprintf( logfp, "%s", outbuf );
      break;
    case CON:
      fprintf( logfp, "SurfAgent: CON: %ld: %s", this_thread, outbuf );
      break;
    case GPC:
      fprintf( logfp, "SurfAgent: GPC: %ld: %s", this_thread, outbuf );
      break;
    case THR:
      fprintf( logfp, "SurfAgent: THR: %ld: %s", this_thread, outbuf );
      break;
    case ERR:
      fprintf( logfp, "SurfAgent: ERR: %ld: %s", this_thread, outbuf );
      break;
    case ERN:
      fprintf( logfp, "SurfAgent: ERN: %ld: %s: %s",
                      this_thread, strerror(errno), outbuf );
      break;
    default:
      fprintf( logfp, "SurfAgent: ERR: %ld: errlogf(): unknown loglevel %d\n"
                      "Current log buffer is: %s\nExiting...\n\n",
                      this_thread, loglevel, outbuf );
      exit( 1 );
      break;
  }

} /* errlogf */


/*----------------------------------------------------------------------------*/
/* write with added logging capabilities (- do we really need this ?)         */
/* we _rely_ on the fact that if n == 0, the buf can be anything, even NULL ! */
/*----------------------------------------------------------------------------*/
int
write_and_log( int fd, char *buf, int n ) {

  if( n <= 0 )
    return( 0 );

  errlogf( LOG, buf, n );

  return( write_socket(fd, buf, n) );

} /* write_and_log */

