/* This file is a component of the JunkBusterMT Proxy, based on original code from
 * the Internet Junkbuster Proxy.
 *
 * Copyright (C) 2000 Gabriel L. Somlo
 * Copyright (C) 1997 Anonymous Coders and Junkbusters Corporation
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see ../GPL.txt); if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <ctype.h>

#ifndef WIN_GCC
#include <netinet/in.h>	/* for the aclfile (ntohl) */
#endif

#include <pthread.h>

#include "filegard.h"

#include "gnu_regex.h"

#include "parsers.h"

#include "filters.h"

#include "urlcodec.h"

#include "errlog.h"
#include "miscstr.h"

#include "bindconn.h"
#include "socks4.h"

#include "blckstat.h"	/* get_url_stats() */

#include "chat.h"	/* things like VERSION, BANNER, etc. */


char *suppress_message = NULL;	/* if set from main(), suppresses listing of
				 * file contents by get_proxy_args() on a
                                 * show-proxy-args page.
                                 */

/*-------------------------------------------------------------*/
/* used by all of acl, block, cookie and forward classes below */
/*-------------------------------------------------------------*/
static void
unload_url( struct url_spec *url ) {

  if( url ) {
    freez( url->spec );
    freez( url->domain );
    freez( url->dbuf );
    freez( url->dvec );
    freez( url->path );
    if( url->preg ) {
      regfree( url->preg );
      freez( url->preg );
    }
  }

} /* unload_url */


/*--------------------------------- aclfile ----------------------------------*/

/* private type, not needed outside this file */
struct access_control_list {
	struct access_control_addr src[1];
	struct access_control_addr dst[1];

	short	action;
	struct access_control_list *next;
};

static char                       *aclfile = NULL;
static struct stat                aclstat[1];
static pthread_mutex_t            aclfile_mutex;
static struct access_control_list *acl = NULL;
static pthread_rwlock_t           acl_lock;

#define ACL_PERMIT	0	/* accept connection request */
#define ACL_DENY  	1	/* reject connection request */

/*------------------------*/
/* used by load_aclfile() */
/*------------------------*/
static int
acl_addr( char *aspec, struct access_control_addr *aca ) {

  int i, masklength, port;
  char *p;

  masklength = 32;
  port       =  0;

  if( (p = strchr(aspec, '/')) ) {
    *p++ = '\0';

    if( !isdigit((int)(*p)) )
      return( -1 );

    masklength = atoi( p );
  }

  if( (masklength < 0) || (masklength > 32) )
    return( -1 );

  if( (p = strchr(aspec, ':')) ) {
    *p++ = '\0';

    if( !isdigit((int)(*p)) )
      return(-1);

    port = atoi( p );
  }

  aca->port = port;

  aca->addr = ntohl( atoip(aspec) );

  if( aca->addr == -1 ) {
    errlogf( ERN, "acl_addr(): can't resolve address for %s\n", aspec);
    return( -1 );
  }

  /* build the netmask */
  for( aca->mask = 0, i = 1; i <= masklength; ++i )
    aca->mask |= ( 1 << (32 - i) );


  /* now mask off the host portion of the ip address
   * (i.e. save on the network portion of the address).
   */
  aca->addr = aca->addr & aca->mask;

  return( 0 );

} /* acl_addr */

/*-----------------------------------------------------*/
/* assumes acl is enabled (aclfile is set to non-null) */
/* checks if acl is current, and reloads it if not     */
/* returns 0 if ok, -1 on error                        */
/*-----------------------------------------------------*/
static int
load_aclfile( void ) {

  FILE *fp;

  char buf[BUFSIZ], *v[3], *p;
  int i;
  struct access_control_list *a, *bl;
  struct stat curr[1];

  pthread_mutex_lock( &aclfile_mutex );

  if( stat(aclfile, curr) < 0 ) {
    errlogf( ERN, "load_aclfile(): error getting stat on aclfile: %s\n",
                  aclfile );
    pthread_mutex_unlock( &aclfile_mutex );
    return( -1 );
  }

  if( acl && (aclstat->st_mtime == curr->st_mtime) ) {
    /* file unchanged, keep using current acl */
    pthread_mutex_unlock( &aclfile_mutex );
    return( 0 );
  }

  if( !(bl = zalloc(sizeof(struct access_control_list))) ) {
    errlogf( ERN, "load_aclfile(): bl = zalloc(...) failed !\n" );
    pthread_mutex_unlock( &aclfile_mutex );
    return( -1 );
  }

  *aclstat = *curr;

  if( !(fp = fopen(aclfile, "r")) ) {
    errlogf( ERN, "load_aclfile(): fopen failed on aclfile: %s\n", aclfile );
    pthread_mutex_unlock( &aclfile_mutex );
    return( -1 );
  }

  while( fgets(buf, sizeof(buf), fp) ) {

    if( (p = strpbrk(buf, "#\r\n")) )
      *p = '\0';

    if( *buf == '\0' )
      continue;

    i = ssplit(buf, " \t", v, SZ(v), 1, 1);

    /* allocate a new node */
    if( !(a = zalloc(sizeof(struct access_control_list))) ) {
      errlogf( ERN, "load_aclfile(): a = zalloc(...) failed !\n" );
      fclose( fp );
      pthread_mutex_unlock( &aclfile_mutex );
      return( -1 );
    }

    /* add it to the list */
    a->next = bl->next;
    bl->next = a;

    switch( i ) {
      case 3:
        if( acl_addr(v[2], a->dst) < 0 ) {
          errlogf( ERR, "load_aclfile(): error in switch(...) case 3\n" );
          pthread_mutex_unlock( &aclfile_mutex );
          return( -1 );
        }
        /* no break */
      case 2:
        if( acl_addr(v[1], a->src) < 0 ) {
          errlogf( ERR, "load_aclfile(): error in switch(...) case 2\n" );
          pthread_mutex_unlock( &aclfile_mutex );
          return( -1 );
        }

        p = v[0];
        if( strcasecmp(p, "permit") == 0 ) {
          a->action = ACL_PERMIT;
          break;
        }

        if( strcasecmp(p, "deny") == 0 ) {
          a->action = ACL_DENY;
          break;
        }
        /* no break */
      default:
        errlogf( ERR, "load_aclfile(): error in switch(...) case default\n" );
        pthread_mutex_unlock( &aclfile_mutex );
        return( -1 );
    } /* switch( i ) */

  } /* while( fgets(buf, ..., fp) ) */

  fclose( fp );

  /* now replace the old acl with the new */
  pthread_rwlock_wrlock( &acl_lock );
  a = acl;			/* keep old one to destroy */
  acl = bl;			/* set the new version */
  pthread_rwlock_unlock( &acl_lock );

  pthread_mutex_unlock( &aclfile_mutex );

  /*unload_aclfile( a );*/	/* we can take our time to delete the old one */
  while( (bl = a) ) {
    a = bl->next;

    free( bl );
  }

  return( 0 );

} /* load_aclfile */

/*-------------------------------------------------------*/
/* initializes use of acl by setting aclfile to non-null */
/* also, initializes locks if necessary                  */
/*-------------------------------------------------------*/
/* public */
void
init_acl( char *aclfname ) {

  if( aclfname ) {
    aclfile = strdup( aclfname );
    pthread_rwlock_init( &acl_lock, NULL );
    pthread_mutex_init( &aclfile_mutex, NULL );
  }

} /* init_acl */

/*-------------------------------------------------------------------------*/
/* main acl function: runs a check on whether the src->dst pair is allowed */
/* will call load_aclfile() to make sure the acl list is current           */
/* then it will read_lock acl and perform the check                        */
/*-------------------------------------------------------------------------*/
/* public */
int
block_acl( struct access_control_addr *src, struct access_control_addr *dst ) {

  struct access_control_list *a;
  struct access_control_addr s[1], d[1];
  int retval;

  /* if not using an access control list, then permit the connection    */
  /* the object is uninitialized, messing with locks makes no sense now */
  if( !aclfile )
    return( ACL_PERMIT );

  if( load_aclfile() != 0 ) {
    errlogf( ERR, "block_acl(): error from load_aclfile() call. Exiting...\n" );
    exit( 1 );
  }

  pthread_rwlock_rdlock( &acl_lock );

  /* if access control list empty, warn (should not really ever happen...) */
  if( !acl ) {
    errlogf( ERR, "block_acl(): WARNING: acl is NULL. Allowing all !!!\n" );
    pthread_rwlock_unlock( &acl_lock );
    return( ACL_PERMIT );
  }

  /* search the list */
  for( a = acl->next; a; a = a->next ) {

    *s = *src;
    *d = *dst;

    s->addr &= a->src->mask;
    d->addr &= a->dst->mask;

    if( (s->addr == a->src->addr) &&
        (d->addr == a->dst->addr) &&
        ((s->port == a->src->port) || (s->port == 0) || (a->src->port == 0)) &&
        ((d->port == a->dst->port) || (d->port == 0) || (a->dst->port == 0)) ) {

      retval = a->action;

      pthread_rwlock_unlock( &acl_lock );
      return( retval );
    }
  }

  /* not found on the list - I guess the default is to deny... */
  pthread_rwlock_unlock( &acl_lock );
  return( ACL_DENY );

} /* block_acl */

/*----------------------------- end aclfile ----------------------------------*/


/*------------------------------- blockfile ----------------------------------*/

/* private type, not needed outside this file */
struct block_spec {
	struct url_spec url[1];
	int   reject;
	struct block_spec *next;
};

static char              *blockfile = NULL;
static struct stat       blockstat[1];
static pthread_mutex_t   blockfile_mutex;
static struct block_spec *block = NULL;
static char              *blockargs = NULL;
static pthread_rwlock_t  block_lock;

/*--------------------------------------------------*/
/* assumes block is enabled (blockfile is non-null) */
/* checks if block is current, and reloads if not   */
/* returns 0 if ok, -1 on error                     */
/*--------------------------------------------------*/
static int
load_blockfile( void ) {

  FILE *fp;

  struct block_spec *b, *bl;
  char  buf[BUFSIZ], *p, *q, *new_blockargs = NULL;
  int port, reject;
  struct stat curr[1];
  struct url_spec url[1];

  pthread_mutex_lock( &blockfile_mutex );

  if( stat(blockfile, curr) < 0 ) {
    errlogf( ERN, "load_blockfile(): error getting stat on blockfile: %s\n",
                  blockfile );
    pthread_mutex_unlock( &blockfile_mutex );
    return( -1 );
  }

  if( block && (blockstat->st_mtime == curr->st_mtime) ) {
    /* file unchanged, keep using current block */
    pthread_mutex_unlock( &blockfile_mutex );
    return( 0 );
  }

  if( !(bl = zalloc(sizeof(struct block_spec))) ) {
    errlogf( ERN, "load_blockfile(): bl = zalloc(...) failed !\n" );
    pthread_mutex_unlock( &blockfile_mutex );
    return( -1 );
  }

  *blockstat = *curr;

  if( !(fp = fopen(blockfile, "r")) ) {
    errlogf( ERN, "load_blockfile(): fopen failed on blockfile: %s\n",
                  blockfile );
    pthread_mutex_unlock( &blockfile_mutex );
    return( -1 );
  }

  if( !suppress_message ) {
    /* initialize argument string */
    p = html_encode( blockfile );
    sprintf( buf, "<h2>The file `%s' contains the following "
                  "patterns</h2>\n<pre>",
                  p );
    freez( p );
    new_blockargs = strsav( new_blockargs, buf );
  }

  while( fgets(buf, sizeof(buf), fp) ) {

    if( !suppress_message ) {
      /* add buf to new_blockargs */
      if( (p = html_encode(buf)) ) {
        new_blockargs = strsav( new_blockargs, p );
        new_blockargs = strsav( new_blockargs, "<br>" );
        free( p );
      }
    }

    if( (p = strpbrk(buf, "\r\n")) )
      *p = '\0';

    /* comments */
    if( (p = strchr(buf, '#')) )
      *p = '\0';

    /* elide white-space */
    for( p = q = buf; *q ; q++ )
      if( !isspace((int)(*q)) )
        *p++ = *q;

    *p = '\0';

    reject = 1;

    if( *buf == '~' ) {
      reject = 0;
      p = buf;
      q = p + 1;
      while( (*p++ = *q++) ); /* what the hell is THIS doing ? Ah! shifting the
                               * buffer left over the '~' after having set
                               * _reject_ to 0 !!!
                               */
    }

    /* skip blank lines */
    if( *buf == '\0' )
      continue;

    /* allocate a new node */
    if( !(b = zalloc(sizeof(struct block_spec))) ||
        !(b->url->preg = zalloc(sizeof(regex_t))) ) {
      errlogf( ERN, "load_blockfile(): zalloc failed to create new node !\n" );
      fclose( fp );
      pthread_mutex_unlock( &blockfile_mutex );
      return( -1 );
    }

    /* add it to the list */
    b->next = bl->next;
    bl->next = b;

    /* save a copy of the orignal specification */
    if( !(b->url->spec = strdup(buf)) ) {
      errlogf( ERN, "load_blockfile(): invalid original specification !\n" );
      fclose( fp );
      pthread_mutex_unlock( &blockfile_mutex );
      return( -1 );
    }

    b->reject = reject;

    if( (p = strchr(buf, '/')) ) {
      b->url->path = strdup( p );
      b->url->pathlen = strlen( b->url->path );
      *p = '\0';
    } else {
      b->url->path = NULL;
      b->url->pathlen = 0;
    }

    if( b->url->path ) {
      int errcode;
      char rebuf[BUFSIZ];

      sprintf( rebuf, "^(%s)", b->url->path );

      errcode = regcomp( b->url->preg, rebuf,
                         REG_EXTENDED|REG_NOSUB|REG_ICASE );

      if( errcode ) {
        size_t errlen = regerror( errcode, b->url->preg, buf, sizeof(buf) );

        buf[errlen] = '\0';

        errlogf( ERR, "load_blockfile(): error compiling %s: %s\n",
                      b->url->spec, buf );
        fclose( fp );
        pthread_mutex_unlock( &blockfile_mutex );
        return( -1 );
      }
    } else {
      freez( b->url->preg );
    }

    if( !(p = strchr(buf, ':')) ) {
      port = 0;
    } else {
      *p++ = '\0';
      port = atoi( p );
    }

    b->url->port = port;

    if( !(b->url->domain = strdup(buf)) ) {
      errlogf( ERN, "load_blockfile(): invalid url domain !\n" );
      fclose( fp );
      pthread_mutex_unlock( &blockfile_mutex );
      return( -1 );
    }

    /* split domain into components */
    *url = dsplit( b->url->domain );
    b->url->dbuf = url->dbuf;
    b->url->dcnt = url->dcnt;
    b->url->dvec = url->dvec;

  } /* while( fgets(buf, ..., fp) ) */

  if( !suppress_message ) {
    /* terminate new_blockargs string */
    new_blockargs = strsav( new_blockargs, "</pre>" );
  }

  fclose( fp );

  /* now replace the old block with the new */
  pthread_rwlock_wrlock( &block_lock );
  b = block;
  block = bl;
  p = blockargs;
  blockargs = new_blockargs;
  pthread_rwlock_unlock( &block_lock );

  pthread_mutex_unlock( &blockfile_mutex );

  /* remove old blockargs (now p) and old block (now b) */
  freez( p );
  while( (bl = b) ) {
    b = bl->next;
    unload_url( bl->url );
    free( bl );
  }

  return( 0 );

} /* load_blockfile */

/*---------------------------------------------------------------*/
/* initializes the use of block by setting blockfile to non-null */
/* also, initializes locks if necessary                          */
/*---------------------------------------------------------------*/
void
init_block( char *blockfname ) {

  if( blockfname ) {
    blockfile = strdup( blockfname );
    pthread_rwlock_init( &block_lock, NULL );
    pthread_mutex_init( &blockfile_mutex, NULL );
  }

} /* init_block */

/*-------------------------------------------------------------------------*/
/* main block function: runs a check on whether the current url is allowed */
/* will call load_blockfile() to make sure the block list is current       */
/* then it will read_lock block and perform the check                      */
/*-------------------------------------------------------------------------*/
char *
is_blocked_url( struct http_request *http ) {

  struct block_spec *b;
  struct url_spec url[1];
  char *retval;

  /* if not using a blockfile, then permit the connection               */
  /* the object is uninitialized, messing with locks makes no sense now */
  if( !blockfile )
    return( NULL );

  if( load_blockfile() != 0 ) {
    errlogf( ERR, "is_blocked_url(): error from load_blockfile() call. "
                  "Exiting...\n" );
    exit( 1 );
  }

  pthread_rwlock_rdlock( &block_lock );

  /* if block list empty, warn */
  if( !block ) {
    errlogf( ERR, "is_blocked_url(): WARNING: block is NULL. "
                  "Allowing all !!!\n" );
    pthread_rwlock_unlock( &block_lock );
    return( NULL );
  }

  *url = dsplit( http->host );

  /* if splitting the domain fails, punt */
  if( !(url->dbuf) ) {
    pthread_rwlock_unlock( &block_lock );
    return( NULL );
  }

  for( b = block->next; b ; b = b->next )
    if( ((b->url->port == 0) || (b->url->port == http->port)) &&
        ((b->url->domain[0] == '\0') || (domaincmp(b->url, url) == 0)) &&
        ((b->url->path == NULL) ||
         (regexec(b->url->preg, http->path, 0, NULL, 0) == 0))
      ) {
      freez( url->dbuf );
      freez( url->dvec );
      if( b->reject ) {
        retval = strdup( b->url->spec );
      } else {
        retval = NULL;
      }

      pthread_rwlock_unlock( &block_lock );
      return( retval );
    }

  freez( url->dbuf );
  freez( url->dvec );

  pthread_rwlock_unlock( &block_lock );
  return( NULL );

} /* is_blocked_url */

/*--------------------------- end blockfile ----------------------------------*/


/*------------------------------ cookiefile ----------------------------------*/

/* private type, not needed outside this file */
struct cookie_spec {
  struct url_spec url[1];
  int cookie_status;
  struct cookie_spec *next;
};

static char               *cookiefile = NULL;
static struct stat        cookiestat[1];
static pthread_mutex_t    cookiefile_mutex;
static struct cookie_spec *cookieblock = NULL;
static char               *cookieargs = NULL;
static pthread_rwlock_t   cookieblock_lock;

/*---------------------------------------------------------------------*/
/* assumes cookie filtering is enabled (cookiefile is set to non-null) */
/* checks if cookieblock is current, and reloads it if not             */
/* returns 0 if ok, -1 on error                                        */
/*---------------------------------------------------------------------*/
static int
load_cookiefile( void ) {

  FILE *fp;

  struct cookie_spec *b, *bl;
  char  buf[BUFSIZ], *p, *q, *new_cookieargs = NULL;
  int port, status;
  struct stat curr[1];
  struct url_spec url[1];

  pthread_mutex_lock( &cookiefile_mutex );

  if( stat(cookiefile, curr) < 0 ) {
    errlogf( ERN, "load_cookiefile(): error getting stat on cookiefile: %s\n",
                  cookiefile );
    pthread_mutex_unlock( &cookiefile_mutex );
    return( -1 );
  }

  if( cookieblock && (cookiestat->st_mtime == curr->st_mtime) ) {
    /* file unchanged, keep using current cookieblock */
    pthread_mutex_unlock( &cookiefile_mutex );
    return( 0 );
  }

  if( !(bl = zalloc(sizeof(struct cookie_spec))) ) {
    errlogf( ERN, "load_cookiefile(): bl = zalloc(...) failed !\n" );
    pthread_mutex_unlock( &cookiefile_mutex );
    return( -1 );
  }

  *cookiestat = *curr;
	
  if( !(fp = fopen(cookiefile, "r")) ) {
    errlogf( ERN, "load_cookiefile(): fopen failed on cookiefile: %s\n",
                  cookiefile );
    pthread_mutex_unlock( &cookiefile_mutex );
    return( -1 );
  }

  if( !suppress_message ) {
    /* initialize argument string */
    p = html_encode( cookiefile );
    sprintf( buf, "<h2>The file `%s' contains the following "
                  "patterns</h2>\n<pre>",
                  p );
    freez( p );
    new_cookieargs = strsav( new_cookieargs, buf );
  }

  while( fgets(buf, sizeof(buf), fp) ) {

    if( !suppress_message ) {
      /* add buf to new_cookieargs */
      if( (p = html_encode(buf)) ) {
        new_cookieargs = strsav( new_cookieargs, p );
        new_cookieargs = strsav( new_cookieargs, "<br>" );
        free( p );
      }
    }

    if( (p = strpbrk(buf, "\r\n")) )
      *p = '\0';

    /* comments */
    if( (p = strchr(buf, '#')) )
      *p = '\0';

    /* elide white-space */
    for( p = q = buf; *q ; q++ )
      if( !isspace((int)(*q)) )
        *p++ = *q;

    *p = '\0';

    p = buf;

    switch( *p ) {
      case '>':
        status = COOKIE_MSK_SND_CLI;	/* allow outgoing */
        p++;
        break;
      case '<':
        status = COOKIE_MSK_ACC_SRV;	/* allow incoming */
        p++;
        break;
      case '~':
        status = 0;			/* deny */
        p++;
        break;
      case '$':
        status = COOKIE_MSK_SMART;	/* store on proxy */
        p++;
        break;
      default:
        status = COOKIE_MSK_ACC_SRV | COOKIE_MSK_SND_CLI;	/* allow */
        break;
    }

    /* elide any of the "special" chars from the front of the pattern */
    q = buf;
    if( p > q )
      while( (*q++ = *p++) );

    /* skip blank lines */
    if( *buf == '\0' )
      continue;

    /* allocate a new node */
    if( !(b = zalloc(sizeof(struct cookie_spec))) ||
        !(b->url->preg = zalloc(sizeof(*b->url->preg))) ) {
      errlogf( ERN, "load_cookiefile(): zalloc failed to create new node !\n" );
      fclose( fp );
      pthread_mutex_unlock( &cookiefile_mutex );
      return( -1 );
    }

    /* add it to the list */
    b->next = bl->next;
    bl->next = b;

    /* save a copy of the orignal specification */
    if( !(b->url->spec = strdup(buf)) ) {
      errlogf( ERN, "load_cookiefile(): invalid original specification !\n" );
      fclose( fp );
      pthread_mutex_unlock( &cookiefile_mutex );
      return( -1 );
    }

    b->cookie_status = status;

    if( (p = strchr(buf, '/')) ) {
      b->url->path = strdup( p );
      b->url->pathlen = strlen( b->url->path );
      *p = '\0';
    } else {
      b->url->path = NULL;
      b->url->pathlen = 0;
    }

    if( b->url->path ) {
      int errcode;
      char rebuf[BUFSIZ];

      sprintf( rebuf, "^(%s)", b->url->path );

      errcode = regcomp( b->url->preg, rebuf,
                         REG_EXTENDED|REG_NOSUB|REG_ICASE );
      if(errcode) {
        size_t errlen = regerror( errcode, b->url->preg, buf, sizeof(buf) );

        buf[errlen] = '\0';

        errlogf( ERR, "load_cookiefile(): error compiling %s: %s\n",
                      b->url->spec, buf );
        fclose( fp );
        pthread_mutex_unlock( &cookiefile_mutex );
        return( -1 );
      }
    } else {
      freez(b->url->preg);
    }

    if( !(p = strchr(buf, ':')) ) {
      port = 0;
    } else {
      *p++ = '\0';
      port = atoi(p);
    }

    b->url->port = port;

    if( !(b->url->domain = strdup(buf)) ) {
      errlogf( ERN, "load_cookiefile(): invalid url domain !\n" );
      fclose( fp );
      pthread_mutex_unlock( &cookiefile_mutex );
      return( -1 );
    }

    /* split domain into components */
    *url = dsplit( b->url->domain );
    b->url->dbuf = url->dbuf;
    b->url->dcnt = url->dcnt;
    b->url->dvec = url->dvec;

  } /* while( fgets(buf, ..., fp) ) */

  if( !suppress_message ) {
    /* terminate new_cookieargs */
    new_cookieargs = strsav( new_cookieargs, "</pre>" );
  }

  fclose( fp );

  /* now replace the old cookieblock with the new */
  pthread_rwlock_wrlock( &cookieblock_lock );
  b = cookieblock;
  cookieblock = bl;
  p = cookieargs;
  cookieargs = new_cookieargs;
  pthread_rwlock_unlock( &cookieblock_lock );

  pthread_mutex_unlock( &cookiefile_mutex );

  /* remove old cookieargs (now p) and old cookieblock (now b) */
  freez( p );
  while( (bl = b) ) {
    b = bl->next;

    unload_url( bl->url );
    free( bl );
  }

  return( 0 );

} /* load_cookiefile */

/*---------------------------------------------------------------------*/
/* initialize the use of cookieblock by setting cookiefile to non-null */
/* also, initializes locks if necessary                                */
/*---------------------------------------------------------------------*/
void
init_cookieblock( char *cookiefname ) {

  if( cookiefname ) {
    cookiefile = strdup( cookiefname );
    pthread_rwlock_init( &cookieblock_lock, NULL );
    pthread_mutex_init( &cookiefile_mutex, NULL );
  }

} /* init_cookieblock */

/*--------------------------------------------------------------------------*/
/* main cookieblock function: runs a check on the cookie status of the      */
/*  current request                                                         */
/* will call load_cookiefile() to make sure the cookieblock list is current */
/* then it will read_lock cookieblock and perform the check                 */
/*--------------------------------------------------------------------------*/
int
cookie_url( struct http_request *http ) {

  struct cookie_spec *b;
  struct url_spec url[1];
  int retval;

  /* if not using a cookiefile, then permit everything                     */
  /* if the object is uninitialized, messing with locks makes no sense now */
  if( !cookiefile )
    return( COOKIE_MSK_SND_CLI | COOKIE_MSK_ACC_SRV );

  if( load_cookiefile() != 0 ) {
    errlogf( ERR, "cookie_url(): error from load_cookiefile() call. "
                  "Exiting...\n" );
    exit( 1 );
  }

  pthread_rwlock_rdlock( &cookieblock_lock );

  /* if cookieblock list empty, warn and deny */
  if( !cookieblock ) {
    errlogf( ERR, "cookie_url(): WARNING: cookieblock NULL. Denying all !!\n" );
    pthread_rwlock_unlock( &cookieblock_lock );
    return( 0 );
  }

  *url = dsplit( http->host );

  /* if splitting the domain fails, deny */
  if( !(url->dbuf) ) {
    pthread_rwlock_unlock( &cookieblock_lock );
    return( 0 );
  }

  for( b = cookieblock->next; b ; b = b->next)
    if( ((b->url->port == 0) || (b->url->port == http->port)) &&
        ((b->url->domain[0] == '\0') || (domaincmp(b->url, url) == 0)) &&
        ((b->url->path == NULL) ||
         (regexec(b->url->preg, http->path, 0, NULL, 0) == 0))
      ) {
      freez( url->dbuf );
      freez( url->dvec );

      retval = b->cookie_status;

      pthread_rwlock_unlock( &cookieblock_lock );
      return( retval );
    }

  freez( url->dbuf );
  freez( url->dvec );

  pthread_rwlock_unlock( &cookieblock_lock );
  return( 0 );

} /* cookie_url */

/*-------------------------- end cookiefile ----------------------------------*/


/*----------------------------- forwardfile ----------------------------------*/

/* private type, not needed outside this file */
struct forward_spec {
	struct url_spec url[1];
	int   reject;
	struct gateway gw[1];
	struct forward_spec *next;
};

static char                *forwardfile = NULL;
static struct stat         fwdstat[1];
static pthread_mutex_t     forwardfile_mutex;
static struct forward_spec *fwd = NULL;
static char                *fwdargs = NULL;
static pthread_rwlock_t    fwd_lock;

static struct gateway gateways[] = {
  /* name      conn. function  type      gwhos gwpt,  fwhos fwpt */
  { "direct",  direct_connect, 0,        NULL, 0,     NULL, 0    },
  { ".",       direct_connect, 0,        NULL, 0,     NULL, 0    },
  { "socks",   socks4_connect, SOCKS_4,  NULL, 1080,  NULL, 0    },
  { "socks4",  socks4_connect, SOCKS_4,  NULL, 1080,  NULL, 0    },
  { "socks4a", socks4_connect, SOCKS_4A, NULL, 1080,  NULL, 0    },
  { NULL,      NULL,           0,        NULL, 0,     NULL, 0    }
};
static struct gateway *gw_default = &(gateways[0]);

/*--------------------------------------------------*/
/* assumes fwd is enabled (forwardfile is non-null) */
/* checks if fwd is current, and reloads if not     */
/* returns 0 if ok, -1 on error                     */
/*--------------------------------------------------*/
static int
load_forwardfile( void ) {

  FILE *fp;

  struct forward_spec *b, *bl;
  char  buf[BUFSIZ], *p, *q, *tmp, *new_fwdargs = NULL;
  char  *vec[4];
  int port, reject;
  struct gateway *gw;
  struct stat curr[1];
  struct url_spec url[1];

  pthread_mutex_lock( &forwardfile_mutex );

  if( stat(forwardfile, curr) < 0 ) {
    errlogf( ERN, "load_forwardfile(): error getting stat on forwardfile: %s\n",
                  forwardfile );
    pthread_mutex_unlock( &forwardfile_mutex );
    return( -1 );
  }

  if( fwd && (fwdstat->st_mtime == curr->st_mtime) ) {
    /* file unchanged, keep using current fwd */
    pthread_mutex_unlock( &forwardfile_mutex );
    return( 0 );
  }

  if( !(bl = zalloc(sizeof(struct forward_spec))) ) {
    errlogf( ERN, "load_forwardfile(): bl = zalloc(...) failed !\n" );
    pthread_mutex_unlock( &forwardfile_mutex );
    return( -1 );
  }

  *fwdstat = *curr;

  if( !(fp = fopen(forwardfile, "r")) ) {
    errlogf( ERN, "load_forwardfile(): fopen failed on forwardfile: %s\n",
                  forwardfile );
    pthread_mutex_unlock( &forwardfile_mutex );
    return( -1 );
  }

  if( !suppress_message ) {
    /* initialize argument string */
    p = html_encode( forwardfile );
    sprintf( buf, "<h2>The file `%s' contains the following "
                  "patterns</h2>\n<pre>",
                  p );
    freez( p );
    new_fwdargs = strsav( new_fwdargs, buf );
  }

  tmp = NULL;
  while( fgets(buf, sizeof(buf), fp) ) {

    freez( tmp );

    if( !suppress_message ) {
      /* add buf to new_fwdargs */
      if( (p = html_encode(buf)) ) {
        new_fwdargs = strsav( new_fwdargs, p );
        new_fwdargs = strsav( new_fwdargs, "<br>" );
        freez( p );
      }
    }

    if( (p = strpbrk(buf, "\r\n")) )
      *p = '\0';

    /* comments */
    if( (p = strchr(buf, '#')) )
      *p = '\0';

    /* skip blank lines */
    if( *buf == '\0' )
      continue;

    tmp = strdup( buf );

    if( ssplit(tmp, " \t", vec, SZ(vec), 1, 1) != 4 ) {
      errlogf( ERR, "load_forwardfile(): error in forwardfile: %s\n", buf );
      continue;
    }

    strcpy( buf, vec[0] );

    reject = 1;

    if( *buf == '~' ) {
      reject = 0;
      p = buf;
      q = p+1;
      while( (*p++ = *q++) );
    }

    /* skip blank lines */
    if( *buf == '\0' )
      continue;

    /* allocate a new node */
    if( !(b = zalloc(sizeof(struct forward_spec))) ||
        !(b->url->preg = zalloc(sizeof(regex_t))) ) {
      errlogf( ERN, "load_forwardfile(): zalloc failed to create node !\n" );
      fclose( fp );
      pthread_mutex_unlock( &forwardfile_mutex );
      return( -1 );
    }

    /* add it to the list */
    b->next = bl->next;
    bl->next = b;

    /* save a copy of the orignal specification */
    if( !(b->url->spec = strdup(buf)) ) {
      errlogf( ERN, "load_forwardfile(): invalid original specification !\n" );
      fclose(fp);
      pthread_mutex_unlock( &forwardfile_mutex );
      return( -1 );
    }

    b->reject = reject;

    if( (p = strchr(buf, '/')) ) {
      b->url->path = strdup( p );
      b->url->pathlen = strlen( b->url->path );
      *p = '\0';
    } else {
      b->url->path = NULL;
      b->url->pathlen = 0;
    }

    if( b->url->path ) {
      int errcode;
      char rebuf[BUFSIZ];

      sprintf( rebuf, "^(%s)", b->url->path );

      errcode = regcomp( b->url->preg, rebuf,
                         REG_EXTENDED|REG_NOSUB|REG_ICASE );

      if( errcode ) {
        size_t errlen = regerror( errcode, b->url->preg, buf, sizeof(buf) );

        buf[errlen] = '\0';

        errlogf( ERR, "load_forwardfile(): error compiling %s: %s\n",
                      b->url->spec, buf );
        fclose( fp );
        pthread_mutex_unlock( &forwardfile_mutex );
        return( -1 );
      }
    } else {
      freez( b->url->preg );
    }

    if( !(p = strchr(buf, ':')) ) {
      port = 0;
    } else {
      *p++ = '\0';
      port = atoi( p );
    }

    b->url->port = port;

    if( !(b->url->domain = strdup(buf)) ) {
      errlogf( ERN, "load_forwardfile(): invalid url domain !\n" );
      fclose( fp );
      pthread_mutex_unlock( &forwardfile_mutex );
      return( -1 );
    }

    /* split domain into components */
    *url = dsplit( b->url->domain );
    b->url->dbuf = url->dbuf;
    b->url->dcnt = url->dcnt;
    b->url->dvec = url->dvec;

    /* now parse the gateway specs */
    p = vec[2];

    for( gw = gateways; gw->name; gw++ )
      if( strcmp(gw->name, p) == 0 )
        break;

    if( !(gw->name) ) {
      errlogf( ERR, "load_forwardfile(): invalid gateway name: %s\n", p );
      pthread_mutex_unlock( &forwardfile_mutex );
      return( -1 );
    }

    /* save this as the gateway type */
    *b->gw = *gw;

    /* now parse the gateway host[:port] spec */
    p = vec[3];

    if( strcmp(p, ".") != 0 ) {
      b->gw->gateway_host = strdup( p );

      if( (p = strchr(b->gw->gateway_host, ':')) ) {
        *p++ = '\0';
        b->gw->gateway_port = atoi( p );
      }

      if( b->gw->gateway_port <= 0 ) {
        errlogf( ERR, "load_forwardfile(): invalid gateway port !\n" );
        pthread_mutex_unlock( &forwardfile_mutex );
        return( -1 );
      }

    }

    /* now parse the forwarding spec */
    p = vec[1];

    if( strcmp(p, ".") != 0 ) {
      b->gw->forward_host = strdup( p );

      if( (p = strchr(b->gw->forward_host, ':')) ) {
        *p++ = '\0';
        b->gw->forward_port = atoi( p );
      }

      if( b->gw->forward_port <= 0 )
        b->gw->forward_port = 8000;
    }

  } /* while( fgets(buf, ..., fp) ) */
  freez( tmp );

  if( !suppress_message ) {
    /* terminate new_fwdargs string */
    new_fwdargs = strsav( new_fwdargs, "</pre>" );
  }

  fclose( fp );

  /* now replace the old fwd with the new */
  pthread_rwlock_wrlock( &fwd_lock );
  b = fwd;
  fwd = bl;
  p = fwdargs;
  fwdargs = new_fwdargs;
  pthread_rwlock_unlock( &fwd_lock );

  pthread_mutex_unlock( &forwardfile_mutex );

  /* remove old fwdargs (now p) and old forward list (now b) */
  freez( p );
  while( (bl = b) ) {
    b = bl->next;

    unload_url( bl->url );
    freez( bl->gw->gateway_host );
    freez( bl->gw->forward_host );
    free( bl );
  }

  return( 0 );

} /* load_forwardfile */

/*---------------------------------------------------------------*/
/* initializes the use of fwd by setting forwardfile to non-null */
/* also, initializes locks if necessary                          */
/*---------------------------------------------------------------*/
void
init_fwd( char *forwardfname ) {

  if( forwardfname ) {
    forwardfile = strdup( forwardfname );
    pthread_rwlock_init( &fwd_lock, NULL );
    pthread_mutex_init( &forwardfile_mutex, NULL );
  }

} /* init_fwd */

/*---------------------------------------------------------------------*/
/* main fwd function: runs a check on whether the current request uses */
/*  forwarding through a gateway                                       */
/* will call load_forwardfile() to make sure the fwd list is current   */
/* then it will read_lock fwd and perform the check                    */
/*---------------------------------------------------------------------*/
struct gateway *
forward_url( struct http_request *http ) {

  struct forward_spec *b;
  struct url_spec url[1];
  struct gateway *retval;

  if( !(retval = malloc(sizeof(struct gateway))) ) {
    errlogf( ERN, "forward_url(): zalloc failed !!\n" );
    exit( 1 );
  }
  /* gateway_host and forward_host will now be set to NULL */
  memcpy( retval, gw_default, sizeof(struct gateway) );	/* default retval */

  /* if not using a forwardfile, use default                            */
  /* the object is uninitialized, messing with locks makes no sense now */
  if( !forwardfile )
    return( retval );

  if( load_forwardfile() != 0 ) {
    errlogf( ERN, "forward_url(): error from load_forwardfile() call. "
                  "Exiting...\n" );
    exit( 1 );
  }

  pthread_rwlock_rdlock( &fwd_lock );

  /* if fwd list empty, warn */
  if( !fwd ) {
    errlogf( ERR, "forward_url(): WARNING: fwd is NULL. Returning default !!!\n" );
    pthread_rwlock_unlock( &fwd_lock );
    return( retval );
  }

  *url = dsplit( http->host );

  /* if splitting the domain fails, punt */
  if( !(url->dbuf) ) {
    pthread_rwlock_unlock( &fwd_lock );
    return( retval );
  }

  for( b = fwd->next; b ; b = b->next)
    if( ((b->url->port == 0) || (b->url->port == http->port)) &&
        ((b->url->domain[0] == '\0') || (domaincmp(b->url, url) == 0)) &&
        ((b->url->path == NULL) ||
         (regexec(b->url->preg, http->path, 0, NULL, 0) == 0))
      ) {
      freez( url->dbuf );
      freez( url->dvec );

      /* copy most of b->gw */
      memcpy( retval, b->gw, sizeof(struct gateway) );

      /* make sure the pointers hanging off of retval are distinct */
      if( b->gw->gateway_host )
        retval->gateway_host = strdup( b->gw->gateway_host );
      if( b->gw->forward_host )
        retval->forward_host = strdup( b->gw->forward_host );

      pthread_rwlock_unlock( &fwd_lock );
      return( retval );
    }

  freez( url->dbuf );
  freez( url->dvec );

  return( retval );

} /* forward_url */

/*------------------------- end forwardfile ----------------------------------*/


/*================================ proxy args ================================*/

#define JBMAN "href=\"http://www.junkbusters.com/ht/en/ijb" VERSION_MAJOR "man.html#"

#define JB_BANNER "<strong>Internet J<small>UNK<i><font bgcolor=\"F8F8F0\" color=\"red\">BUSTER</font></i></small></strong><sup><small>TM</small></sup>"


static char *proxy_args_header;
static char *proxy_args_invocation;
static char *proxy_args_gateways;
static char *proxy_args_trailer;

void
savearg( char *cmd, char *arg ) {

  char *cmd_enc, *arg_enc;
  char buf[BUFSIZ];
  static int one_shot = 1;

  if( one_shot ) {
    one_shot = 0;
    proxy_args_invocation = strsav( proxy_args_invocation,
                                    "<br>\n"
                                    "and the following options were set in "
                                    "the configuration file:<br><br>\n" );
  }

  *buf = '\0';

  if( cmd && *cmd ) {
    if( (cmd_enc = html_encode(cmd)) )
      sprintf( buf, "<a " JBMAN "%s\">%s</a> ", cmd_enc, cmd_enc );
    freez( cmd_enc );
  }

  if( arg && *arg ) { 
    if( (arg_enc = html_encode(arg)) ) {
      if( strncasecmp(arg_enc, "http://", 7) == 0 ) {
        strcat( buf, "<a href=\"" );
        strcat( buf, arg_enc );
        strcat( buf, "\">" );
        strcat( buf, arg_enc );
        strcat( buf, "</a>" );
      } else {
        strcat( buf, arg_enc );
      }
    }
    freez( arg_enc );
  }

  strcat( buf, "<br>\n" );

  proxy_args_invocation = strsav( proxy_args_invocation, buf );

} /* savearg */

void
init_proxy_args( int argc, char **argv ) {

  struct gateway *g;
  int i;

  proxy_args_header = strsav( proxy_args_header,
    "HTTP/1.0 200 OK\n"
    "Server: IJ/" VERSION "\n"
    "Content-type: text/html\n\n"
    "<html>"
    "<head>"
    "<title>SurfAgent Proxy Status</title>"
    "</head>\n"
    BODY
    "<center><h1>" BANNER " Proxy Status</h1></center>\n"
    "<h2>You are using " BANNER " V." VERSION "</h2><p>\n"
    "<h2>The program was invoked as follows</h2>\n"
  );

  for( i=0; i < argc; i++ ) {
    proxy_args_header = strsav( proxy_args_header, argv[i] );
    proxy_args_header = strsav( proxy_args_header, " " );
  }

  proxy_args_header = strsav( proxy_args_header, "<br>\n" );

  proxy_args_gateways = strsav( proxy_args_gateways,
    "<h2>It supports the following gateway protocols:</h2>\n" );

  for( g = gateways; g->name; g++ ) {
    proxy_args_gateways = strsav( proxy_args_gateways, g->name );
    proxy_args_gateways = strsav( proxy_args_gateways, " " );
  }

  proxy_args_gateways = strsav( proxy_args_gateways, "<br>\n") ;

} /* init_proxy_args */

void
end_proxy_args( void ) {

  char *b = NULL;

  b = strsav( b, "<p>This instance of " BANNER " supports "
                 "POSIX regular expressions in the path specs.\n"
  );

  b = strsav(b,
    "<p><hr>\n"
    "<table>\n"
    "<tr><td colspan=3><small><small>The " BANNER " Proxy is developed and "
    "maintained by <a href=\"http://www.cs.colostate.edu/~somlo\">Gabriel L. "
    "Somlo</a>;</small></small></td></tr>\n"
    "<tr><td><small><small>The original code has been borrowed from the "
    "</small></small></td><td bgcolor=\"#F8F8F0\"><small><small>"
    JB_BANNER "</small></small></td><td><small><small>proxy "
    "developed by <a href=\"http://www.junkbusters.com\">Junkbusters Corp.</a>;"
    "</small></small></td></tr>\n"
    "<tr><td colspan=3><small><small>Copying and distribution is permitted "
    "under the <a href=\"http://www.fsf.org/copyleft/gpl.html\">"
    "GNU General Public License</a>;</small></small></td></tr>\n"
    "<tr><td colspan=3><small><small>For questions and comments please e-mail: "
    "<a href=\"mailto:somlo@cs.colostate.edu\">somlo@cs.colostate.edu</a>"
    "</small></small></td></tr>\n"
    "</table>\n"
    "</body></html>\n"
  );

  proxy_args_trailer = b;

} /* end_proxy_args */

char *
get_proxy_args( void ) {

  char *p, *retval = NULL;

  retval = strsav( retval, proxy_args_header );
  retval = strsav( retval, proxy_args_invocation );
  retval = strsav( retval, proxy_args_gateways );
  if( (p = get_url_stats()) ) {
    retval = strsav( retval, p );
    free( p );
  }

  if( suppress_message ) {

    if( suppress_message[0] != '\0' ) {
      retval = strsav( retval, "<h2>File contents:</h2>\n" );
      retval = strsav( retval, suppress_message );
      retval = strsav( retval, "\n" );
    }

  } else {

    /* get blocklist arguments */
    if( blockfile ) {
      pthread_rwlock_rdlock( &block_lock );
      retval = strsav( retval, blockargs );
      pthread_rwlock_unlock( &block_lock );
    }

    /* get cookielist arguments */
    if( cookiefile ) {
      pthread_rwlock_rdlock( &cookieblock_lock );
      retval = strsav( retval, cookieargs );
      pthread_rwlock_unlock( &cookieblock_lock );
    }

    /* get forwardlist arguments */
    if( forwardfile ) {
      pthread_rwlock_rdlock( &fwd_lock );
      retval = strsav( retval, fwdargs );
      pthread_rwlock_unlock( &fwd_lock );
    }

  } /* if( suppress_message ) */

  retval = strsav( retval, proxy_args_trailer );

  return( retval );

} /* get_proxy_args */

/*============================== end proxy args ==============================*/
