/* This file is a component of the JunkBusterMT Proxy, based on original code from
 * the Internet Junkbuster Proxy.
 *
 * Copyright (C) 2000 Gabriel L. Somlo
 * Copyright (C) 1997 Anonymous Coders and Junkbusters Corporation
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see ../GPL.txt); if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "errlog.h"


#define SEP_CHAR  1		/* ssplit() separator character */
#define TERM_CHAR 2		/* ssplit() terminator character */
#define CHUNK_SIZE 256		/* ssplit() ptr. array increment */

/* ssplit() - split a string (in-place) into fields
 *      s = string to split
 *      c = list of characters to be used as field separators
 *          (if NULL, use default separators of space, tab, and newline)
 *      v = vector into which field pointers are placed
 *      n = capacity (number of fields) of the vector
 *      m = flag indicating whether to skip multiple separators (1) or treat
 *          them as an empty string each (0)
 *      l = flag indicating whether to skip leading field separators (1)
 */
int
ssplit( char *s, char *c, char *v[], int n, int m, int l ) {

  char t[256];
  char **x = NULL;
  int xsize = 0;
  unsigned char *p;
  int xi = 0;
  int vi = 0;
  int i;

  if( !s )
    return( -1 );			/* nothing to do ... */

  if( !(p = c) )
    p = " \t\n\r";			/* default field separators */

  /* generate separator and terminator map */
  memset( t, '\0', sizeof(t) );
  while( *p )
    t[*p++] = SEP_CHAR;			/* separator  */
  t['\0'] = t['\n'] = TERM_CHAR;	/* terminator */

  p = s;

  /* skip leading separators if requested */
  if( l )
    while( t[*p] == SEP_CHAR )
      p++;

  xsize = CHUNK_SIZE;
  x = malloc( xsize * sizeof(char *) );

  x[xi++] = p;			/* first pointer is the beginning of string */

  /* first pass:  save pointers to the field separators */
  while( t[*p] != TERM_CHAR )
    if( t[*p] == SEP_CHAR ) {		/* if the char is a separator ... */

      *p++ = '\0';		/* null terminate the substring */

      if( xi == xsize ) {	/* get another chunk */
        xsize += CHUNK_SIZE;
        x = realloc( x, xsize * sizeof(char *) );
      }

      x[xi++] = p;		/* save pointer to beginning of next string */

    } else {

      p++;

    } /* if( b == SEP_CHAR ) */

  *p = '\0';			/* null terminate the substring */

  /* second pass: copy the relevant pointers to the output vector */
  for( i=0; i < xi; i++ ) {

    if( m && (*x[i] == '\0') )
        continue;		/* skip empty strings (used to be separators) */

    if( vi < n ) {
      v[vi++] = x[i];
    } else {
      /* overflow */
      vi = -1;
      break;
    }

  } /* for( ... ) */

  free( x );

  return( vi );

} /* ssplit */


/* strsav() takes a pointer to a string stored in a dynamically allocated
 * buffer and a pointer to a string and returns a pointer to a new dynamically
 * allocated space that contains the concatenation of the two input strings
 * the previous space is free()'d by realloc().
 */
char *
strsav( char *old, char *text_to_append ) {

  int old_len, new_len;
  char *p;

  if( !text_to_append || (*text_to_append == '\0') )
    return( old );

  if( old )
    old_len = strlen( old );
  else
    old_len = 0;

  new_len = old_len + strlen( text_to_append ) + 1;

  if( !(p = realloc(old, new_len)) ) {
    errlogf( ERN, "strsav(): realloc(%d) bytes failed!\n", new_len );
    exit( 1 );
  }

  strcpy( p + old_len, text_to_append );

  return( p );

} /* strsav */


/* the way calloc() ought to be -acjc */
/* I don't know what acjc is talking about, but zalloc basically allocates
 * memory and then zeroes it all out. (iirc, calloc does something else...) */
void *
zalloc( size_t size ) {

  void *ret;

  if( (ret = malloc(size)) )
    memset( ret, '\0', size );

  return( ret );

} /* zalloc */


/*----------------------------------------------------------------------------*/
/* Read a double value from a file (which supposedly contains only the double */
/* value. Returns the double through the second parameter. The return value   */
/* 0 on success, -1 on failure ...                                            */
/*----------------------------------------------------------------------------*/
int
read_double_file( const char *filename, double *valp ) {

  FILE *fp;

  if( !filename ||
      !(fp = fopen(filename, "r")) ||
      (fscanf(fp, "%lf", valp) != 1) ) {
    errlogf( ERN, "read_double_from_file(): failed reading from file: %s\n",
                  (filename) ? filename : "null" );
    return( -1 );
  }

  fclose( fp );
  return( 0 );

} /* read_double_file */


/*----------------------------------------------------------------------------*/
/* Write a double value to a file (overwrite any previously existing value)   */
/* Returns 0 on success, -1 on failure                                        */
/*----------------------------------------------------------------------------*/
int
write_double_file( const char *filename, double val ) {

  FILE *fp;

  if( !filename ||
      !(fp = fopen(filename, "w")) ||
      (fprintf(fp, "%41.20f\n", val) < 0) ) {
    errlogf( ERN, "write_double_to_file(): failed writing to file: %s\n",
                  (filename) ? filename : "null" );

    return( -1 );
  }

  fclose( fp );
  return( 0 );

} /* write_double_file */
