/* This file is a component of the JunkBusterMT Proxy, based on original code from
 * the Internet Junkbuster Proxy.
 *
 * Copyright (C) 2000 Gabriel L. Somlo
 * Copyright (C) 1997 Anonymous Coders and Junkbusters Corporation
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see ../GPL.txt); if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include "parsers.h"

#include "errlog.h"
#include "miscstr.h"


/*=============================== struct list ================================*/

/* h = pointer to list 'dummy' header
 * s = string to add to the list
 */
/* move this to misc.c (gls) */
void
enlist( struct list *h, const char *s ) {

  struct list *n, *l;

  if( (n = malloc(sizeof(struct list))) ) {

    n->str = strdup( s );
    n->next = NULL;

    if( (l = h->last) )
      l->next = n;
    else 
      h->next = n;

    h->last = n;
  }

} /* enlist */

/* move this to misc.c (gls) */
void
destroy_list( struct list *h ) {

  struct list *p, *n;

  for( p = h->next; p ; p = n ) {
    n = p->next;
    freez( p->str );
    freez(p);
  }

  memset( h, '\0', sizeof(struct list) );

} /* destroy_list */

char *
list_to_text( struct list *h ) {

  struct list *p;
  char *ret = NULL;
  char *s;
  int size = 0;

  for( p = h->next; p ; p = p->next )
    if( p->str )
      size += strlen( p->str ) + 2;

  if( !(ret = malloc(size + 1)) )
    return( NULL );

  ret[size] = '\0';

  s = ret;

  for( p = h->next; p ; p = p->next )
    if( p->str ) {
      strcpy( s, p->str );
      s += strlen( s );
      *s++ = '\r';
      *s++ = '\n';
    }

  return( ret );

} /* list_to_text */

/*============================= end struct list ==============================*/


/*=========================== struct http_request ============================*/

char *passageword = "|noblock";

void
free_http_request( struct http_request *http ) {

  freez( http->cmd );
  freez( http->gpc );
  freez( http->protocol );
  freez( http->hostport );
  freez( http->userpwd );
  freez( http->host );
  freez( http->path );
  freez( http->ver );

} /* free_http_request */

/* parse out the info from the http request */
void
parse_http_request( const char *req, struct http_request *http ) {

  char *buf, *v[10], *url, *p, *q;
  int n;

  memset( http, '\0', sizeof(struct http_request) );

  http->cmd = strdup( req );

  buf = strdup( req );

  n = ssplit( buf, " \r\n", v, SZ(v), 1, 1 );

  if( n == 3 ) {

    /* this could be a CONNECT request */
    if( strcasecmp(v[0], "connect") == 0 ) {
      http->ssl      = 1;
      http->gpc      = strdup( v[0] );
      http->hostport = strdup( v[1] );
      http->ver      = strdup( v[2] );
    }

    /* or it could be a GET or a POST */
    if( (strcasecmp(v[0], "get")  == 0) ||
        (strcasecmp(v[0], "head") == 0) ||
        (strcasecmp(v[0], "post") == 0) ) {

      http->ssl      = 0;
      http->gpc      = strdup( v[0] );
      url            =         v[1] ;
      http->ver      = strdup( v[2] );

      if( strncasecmp(url, "http://",  7) == 0 ) {
        url += 7;
        http->protocol = strdup( "http://" );
      } else {
        if( strncasecmp(url, "https://", 8) == 0 ) {
          url += 8;
          http->protocol = strdup( "https://" );
        } else {
          url = NULL;
        }
      }

      if( url && (p = strchr(url, '/')) ) {
        http->path = strdup( p );
        *p = '\0';
        http->hostport = strdup( url );
      }
    }
  }

  freez( buf );

  if( !(http->hostport) ) {
    free_http_request( http );
    return;
  }

  buf = strdup( http->hostport );

  /* if url contains password (e.g., proto://LOGIN:PASSWORD@dom:port/path */
  n = ssplit( buf, "@", v, SZ(v), 1, 1 );
  if( n == 2 ) {
    freez( http->hostport );
    http->hostport = strdup( v[1] );
    http->userpwd = strdup( v[0] );
    freez( buf );
    buf = strdup( http->hostport );
  }

  n = ssplit( buf, ":", v, SZ(v), 1, 1 );

  if( n == 1 ) {
    http->host = strdup( v[0] );
    http->port = 80;
  }

  if( n == 2 ) {
    http->host = strdup( v[0] );
    http->port = atoi( v[1] );
  }

  freez( buf );

  if( !(http->host) )
    free_http_request( http );

  if( !(http->path) )
    http->path = strdup( "" );

  /* check for blocklist circumvention (passageword) */
  if( (p = strstr(http->path, passageword)) ) {

    http->noblock = 1;		/* set noblock flag for this request */

    q = p + strlen( passageword );
    while( (*p++ = *q++) );	/* shift over passageword ( a pw b -> a b ) */

    p = strstr( http->cmd, passageword );	/* also fix cmd */
    q = p + strlen( passageword );
    while( (*p++ = *q++) );
  }

} /* parse_http_request */

/* print the contents of an http structure - for debugging */
void
show_http_request( const struct http_request *http ) {

  printf( "\n\n\n****************** start http request *****************\n\n" );

  if( !http ) {
    printf( "null\n" );
    return;
  }

  printf( "noblock = %d\n", http->noblock );
  printf( "cmd = %s\n", (http->cmd) ? http->cmd : "null" );
  printf( "gpc = %s\n", (http->gpc) ? http->gpc : "null" );
  printf( "protocol = %s\n", (http->protocol) ? http->protocol : "null" );
  printf( "hostport = %s\n", (http->hostport) ? http->hostport : "null" );
  printf( "userpwd = %s\n", (http->userpwd) ? http->userpwd : "null" );
  printf( "host = %s\n", (http->host) ? http->host : "null" );
  printf( "port = %d\n", http->port );
  printf( "path = %s\n", (http->path) ? http->path : "null" );
  printf( "ver = %s\n", (http->ver) ? http->ver : "null" );
  printf( "ssl = %d\n", http->ssl );

  printf( "\n\n****************** end http request ******************\n\n\n" );

} /* show_http_request */

/*========================= end struct http_request ==========================*/


/*================================ struct iob ================================*/

/*----------------------------------------------*/
/* add contents of buf (of length n) to the iob */
/*----------------------------------------------*/
void
add_to_iob( struct iob *iob, char *buf, int n ) {

  int have, need;
  char *p;

  if( n <= 0 )
    return;

  have = iob->eod - iob->cur;
  need = have + n;

  if( !(p = malloc(need + 1)) ) {
    errlogf( ERN, "add_to_iob(): failed to malloc() iob\n" );
    exit( 1 );	/* was return( -1 ); but we're in deep shit by now, so quit */
  }

  if( have > 0 )	/* there is something in the buffer - save it */
    memcpy( p, iob->cur, have );

  /* replace the buffer with the new space */
  freez( iob->buf );
  iob->buf = p;

  /* point to the end of the data */
  p += have;

  /* copy the new data into the iob buffer */
  memcpy( p, buf, n );

  /* point to the end of the data */
  p += n ;

  /* null terminate == cheap insurance */
  *p  = '\0';

  /* set the pointers to the new values */
  iob->cur = iob->buf;
  iob->eod = p;

} /* add_to_iob */

/* this (odd) routine will parse the csp->iob and return one of the following:
 *
 * 1) a pointer to a dynamically allocated string that contains a header line
 * 3) ""    indicating that the end of the iob was reached before finding
 *          a complete header line.
 * 2) NULL  indicating that the end of the header was reached
 */
char *
get_next_header_from_iob( struct iob *iob ) {

  char *p, *q, *ret;

  if( !(iob->cur) || !(p = strchr(iob->cur, '\n')) )
    return( "" );	/* couldn't find a complete header */

  *p = '\0';

  ret = strdup( iob->cur );

  iob->cur = p + 1;

  if( (q = strchr(ret, '\r')) )
    *q = '\0';

  /* is this a blank linke (i.e. the end of the header) ? */
  if( *ret == '\0' ) {
    freez( ret );
    return( NULL );
  }

  return( ret );

} /* get_header */

/*--------------------------------------------------------------*/
/* retrieve a copy what's left over in the iob; then clean      */
/*  returns the copy and places the length in n on success, and */
/*          NULL and places 0 in n on empty or failure          */
/*--------------------------------------------------------------*/
char *
flush_iob( struct iob *iob, int *n ) {

  char *ret;

  /* copy leftovers into ret */
  if( (*n = iob->eod - iob->cur) > 0 ) {
    if( !(ret = malloc(*n)) ) {
      errlogf( ERN, "flush_iob(): malloc failed ! \n" );
      exit( 1 );
    }
    memcpy( ret, iob->cur, *n );
  } else {
    ret = NULL;
    *n = 0;
  }

  /* clean up the iob */
  freez( iob->buf );
  iob->cur = iob->eod = NULL;

  return( ret );

} /* flush_iob */

/*============================== end struct iob ==============================*/
