/* This file is a component of the JunkBusterMT Proxy, based on original code from
 * the Internet Junkbuster Proxy.
 *
 * Copyright (C) 2000 Gabriel L. Somlo
 * Copyright (C) 1997 Anonymous Coders and Junkbusters Corporation
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see ../GPL.txt); if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include "sed.h"

#include "parsers.h"

#include "csp.h"	/* for client_state */

#include "gnu_regex.h"

#include "urlcodec.h"

#include "errlog.h"
#include "miscstr.h"
#include "smrtcook.h"
#include "filegard.h"	/* for the COOKIE_MSK_* mask flags */


/* private type - used by sed() */
struct parsers {
	char *str;
	char  len;
	char *(*parser)( struct parsers *v, char *s, struct client_state *csp );
};

typedef void (*header_adder)( struct client_state *csp );


/* header-line processing functions, and the structures matching them to the
   header lines */

#define DEFAULT_UAGENT	"User-Agent: Mozilla/4.5 [en] (WinNT; I)"

char *referrer = NULL;	/* set in main, used by client_referrer() */
char *uagent   = NULL;	/* set in main, used by client_uagent() */
char *from     = NULL;	/* set in main, used by client_from()  */
FILE *jar = NULL;	/* set in main, used by server_set_cookie() */
int add_forwarded = 0;	/* set in main, used by client_x_forwarded() and
                         *             by client_x_forwarded_adder() */

static char *crumble(                    struct parsers *v, char *s,
                                         struct client_state *csp );
static char *client_referrer(            struct parsers *v, char *s,
                                         struct client_state *csp );
static char *client_uagent(              struct parsers *v, char *s,
                                         struct client_state *csp );
static char *client_ua(                  struct parsers *v, char *s,
                                         struct client_state *csp );
static char *client_x_forwarded(         struct parsers *v, char *s,
                                         struct client_state *csp );
static char *client_from(                struct parsers *v, char *s,
                                         struct client_state *csp );
static char *client_send_cookie(         struct parsers *v, char *s,
                                         struct client_state *csp );
static char *client_proxy_authorization( struct parsers *v, char *s,
                                         struct client_state *csp );
static char *client_accept(              struct parsers *v, char *s,
                                         struct client_state *csp );
static char *server_set_cookie(          struct parsers *v, char *s,
                                         struct client_state *csp );
static char *server_content_type(        struct parsers *v, char *s,
                                         struct client_state *csp );

static struct parsers client_patterns[] = {
  { "referer:", /* sic */    8, client_referrer            },
  { "user-agent:",          11, client_uagent              },
  { "ua-",                   3, client_ua                  },
  { "from:",                 5, client_from                },
  { "cookie:",               7, client_send_cookie         },
  { "x-forwarded-for:",     16, client_x_forwarded         },
  { "proxy-authorization:", 20, client_proxy_authorization },
  { "accept:",               7, client_accept              },
  { "proxy-connection:",    17, crumble                    },
  { "if-modified-since:",   18, crumble                    }, /* temp. (GLS) */
  { NULL,                    0, NULL                       }
};

static struct parsers server_patterns[] = {
  { "set-cookie:",   11, server_set_cookie   },
  { "connection:",   11, crumble             },
  { "content-type:", 13, server_content_type },      /* catch text/html (gls) */
  { NULL,             0, NULL                }
};


/* functions that add lines to http headers, and structures that allow access
   to them */

struct list wafer_list[1];	/* set in main, used by client_cookie_adder() */
struct list xtra_list[1];	/* set in main, used by client_xtra_adder() */

static void client_cookie_adder(      struct client_state *csp );
static void client_xtra_adder(        struct client_state *csp );
static void client_x_forwarded_adder( struct client_state *csp );

static header_adder add_client_headers[] = {
  client_cookie_adder,
  client_x_forwarded_adder,
  client_xtra_adder,
  NULL
};

static header_adder add_server_headers[] = {
  NULL
};


/*---------------*/
/* used by sed() */
/*---------------*/
static struct parsers *
match( char *buf, struct parsers *pats ) {

  struct parsers *v;

  if( buf )
    for( v = pats; v->str ; v++ )
      if( strncasecmp(buf, v->str, v->len) == 0 )
        return( v );

  return( NULL );

} /* match */

/* sed(): add, delete or modify lines in the HTTP header streams.
 * on entry, it receives a linked list of headers space that was
 * allocated dynamically (both the list nodes and the header contents).
 *
 * it returns a single pointer to a fully formed header.
 *
 * as a side effect it frees the space used by the original header lines.
 *
 */
char *
sed( int sc, struct client_state *csp ) {

  struct parsers *pats, *v;
  header_adder *more_headers, *f;
  struct list *p;
  char *hdr;

  switch( sc ) {
    case SED_CLI:
      pats = client_patterns;
      more_headers = add_client_headers;
      break;
    case SED_SRV:
      pats = server_patterns;
      more_headers = add_server_headers;
      break;
    default:
      errlogf( ERR, "sed(): impossible parameter sc. Should never happen\n" );
      exit( 1 );
  }


  for( p = csp->headers->next; p ; p = p->next ) {

    errlogf( HDR, "scan: %s", p->str );

    if( (v = match(p->str, pats)) ) {
      hdr = v->parser( v, p->str, csp );
      freez( p->str );
      p->str = hdr;
    }

    errlogf( HDR, "\n" );

  }

  /* place any additional headers on the csp->headers list */
  for( f = more_headers; *f ; f++ )
    (*f)(csp);

  /* add the blank line at the end of the header */
  enlist( csp->headers, "" );

  hdr = list_to_text( csp->headers );

  return( hdr );

} /* sed */


/* Here begins the family of parser functions that reformat header lines;
 * they are contained in the struct parsers tables client_patterns and
 * server_patterns, and used by sed() per request from chat() */

static char *
crumble( struct parsers *v, char *s, struct client_state *csp ) {

  errlogf( HDR, " crunch!" );

  return( NULL );

} /* crubmle */

static char *
client_referrer( struct parsers *v, char *s, struct client_state *csp ) {

  if( !referrer ) {
    /* crunch referrer line */
    errlogf( HDR, " crunch!" );
    return( NULL );
  }

  if( *referrer == '.' )
    /* pass referrer line unconditionally */
    return( strdup(s) );

  if( *referrer == '@' ) {
    /* pass referrer line only if outgoing cookies are also allowed */
    if( csp->cookie_status & COOKIE_MSK_SND_CLI ) {
      return( strdup(s) );
    } else {
      errlogf( HDR, " crunch!" );
      return( NULL );
    }
  }

  if( *referrer == '' ) {
    /* pass referrer line intact if outgoing cookies allowed, else forge */
    /* (by Andreas S. Oesterhelt <oes.paradis.rhein.de>) */
    if( csp->cookie_status & COOKIE_MSK_SND_CLI ) {
      return( strdup(s) );
    } else {
      errlogf( HDR, " forged!" );
      s = strsav( NULL, "Referer: " );
      s = strsav( s, "http://" );
      s = strsav( s, csp->http->hostport );
      s = strsav( s, "/" );
      return( s );
    }
  }

  errlogf( HDR, " modified" );
  s = strsav( NULL, "Referer: " );
  s = strsav( s, referrer );
  return( s );

} /* client_referrer */

static char *
client_uagent( struct parsers *v, char *s, struct client_state *csp ) {

  if( !uagent ) {
    errlogf( HDR, " default" );
    return( strdup(DEFAULT_UAGENT) );
  }

  if( *uagent == '.' )
    return( strdup(s) );

  if( *uagent == '@' ) {
    if( csp->cookie_status & COOKIE_MSK_SND_CLI ) {
      return( strdup(s) );
    } else {
      errlogf( HDR, " default" );
      return( strdup(DEFAULT_UAGENT) );
    }
  }

  errlogf( HDR, " modified" );

  s = strsav( NULL, "User-Agent: " );
  s = strsav(s, uagent );
  return( s );

} /* client_uagent */

static char *
client_ua( struct parsers *v, char *s, struct client_state *csp ) {

  if( !uagent ) {
    errlogf( HDR, " crunch!" );
    return( NULL );
  }

  if( *uagent == '.' )
    return( strdup(s) );

  if( *uagent == '@' ) {
    if( csp->cookie_status & COOKIE_MSK_SND_CLI ) {
      return( strdup(s) );
    } else {
      errlogf( HDR, " crunch!" );
      return( NULL );
    }
  }

  errlogf( HDR, " crunch!" );

  return( NULL );

} /* client_ua */

static char *
client_from( struct parsers *v, char *s, struct client_state *csp ) {

  /* if not set, zap it */
  if( !from ) {
    errlogf( HDR, " crunch!" );
    return( NULL );
  }

  if( *from == '.' )
    return( strdup(s) );

  errlogf( HDR, " modified" );

  s = strsav( NULL, "From: " );
  s = strsav( s, from );
  return( s );

} /* client_from */

static char *
client_send_cookie( struct parsers *v, char *s, struct client_state *csp ) {

  if( csp->cookie_status & COOKIE_MSK_SND_CLI ) {
    enlist( csp->cookie_list, s + v->len + 1 );
  } else {
    errlogf( HDR, " crunch!" );
    if( csp->cookie_status & COOKIE_MSK_SMART )
      errlogf( HDR, " (SmartCookie)" );
  }

  /* always return NULL here.  the cookie header will be sent
   * at the end of the header.
   */
  return( NULL );

} /* client_send_cookie */

static char *
client_x_forwarded( struct parsers *v, char *s, struct client_state *csp ) {

  if(add_forwarded)
    csp->x_forwarded = strdup( s );

  /* return NULL, since this info will be sent at the end of the header. */
  return( NULL );

} /* client_x_forwarded */

static char *
client_proxy_authorization( struct parsers *v,
                            char *s, struct client_state *csp ) {

  int n;
  char *av[5];

  if( (n = ssplit((s + v->len + 1), " ", av, SZ(av), 1, 1)) != 2 ) {
    errlogf( ERR, "\nclient_proxy_authorization(): Bad \"Proxy-authorization\" field\n" );
    return( NULL );
  }

  /* save user account id in csp */
  csp->user_id = strdup( av[1] );

  errlogf( HDR, " << user acct. id is: %s >>", csp->user_id );

  /* return NULL, since we're not forwarding this to the server... */
  return( NULL );

} /* client_proxy_authorization */

static char *
client_accept( struct parsers *v, char *s, struct client_state *csp ) {

  /* This routine sets csp->accept_image_only; Because http allows for multiple
   * "accept:" header lines, and one non-image in any of them is enough to turn
   * the flag off, the csp has the flag set to "true" by default. This routine
   * then looks for reasons to turn it off...
   */

  int i, n;
  char *intact, *av[20];

  intact = strdup( s );			/* return an intact copy of s */

  if( (n = ssplit((s + v->len + 1), " ,", av, SZ(av), 1, 1)) < 0 ) {
    errlogf( ERR, "\nclient_accept(): av[] array too small !!!\n" );
    csp->accept_image_only = 0;		/* not sure, assume no */
    return( intact );
  }

  for( i = 0; i < n; ++i )
    if( strncasecmp(av[i], "image/", 6) != 0 )
      csp->accept_image_only = 0;

  return( intact );

} /* client_accept */

static char *
server_set_cookie( struct parsers *v, char *s, struct client_state *csp ) {

  if( jar )
    fprintf( jar, "%s\t%s\n", csp->http->host, (s + v->len + 1) );

  if( csp->cookie_status & COOKIE_MSK_ACC_SRV )
    return( strdup(s) );

  if( csp->cookie_status & COOKIE_MSK_SMART ) {

    /* Insert or replace SmartCookie */
    /* should we use hostport instead of host for domain ?
     */
    SmartCookieInsertReplace( csp->http->host,
                              csp->http->path,
                              s + v->len + 1 );

    return( NULL );
  } /* if( csp->cookie_status & COOKIE_MSK_SMART ) */

  /* no accept, no smartcookie  -> crunch */
  return( crumble(v, s, csp) );

} /* server_set_cookie */

/* catch whether document is text/html, or image */
static char *
server_content_type( struct parsers *v, char *s, struct client_state *csp ) {

  char *buf, *comp[5];

  buf = strdup( csp->headers->next->str );	/* copy 1st header line */
  ssplit( buf, " ", comp, SZ(comp), 1, 1 );

  csp->content_is_text_html = 0;

  if( (strncasecmp(s+14, "text/html", 9) == 0) &&
      (strcasecmp(comp[1], "200") == 0)	/* ignore "not found", "moved", etc. */
    ) {
    csp->content_is_text_html = 1;
    errlogf( HDR, " <<< Text or HTML detected !! >>>" );
  }

  freez( buf );

  return( strdup(s) );

} /* server_content_type */


/* Family of functions that add headers directly to the header list (they are
 * contained in the add_client_headers and add_server_headers tables, used
   by sed() per request from chat()
 */

static void
client_cookie_adder( struct client_state *csp ) {

  struct list *l;
  char *tmp = NULL;
  char *e;

  if( csp->cookie_status & COOKIE_MSK_SMART )
    /* retrieve applicable SmartCookies */
    /* maybe we should change host to hostport ?
     */
    tmp = SmartCookieRetrieve( tmp, csp->http->host, csp->http->path );

  for( l = csp->cookie_list->next; l ; l = l->next ) {
    if( tmp )
      tmp = strsav( tmp, "; " );
    tmp = strsav( tmp, l->str );
  }

  for( l = wafer_list->next;  l ; l = l->next ) {
    if( tmp )
      tmp = strsav( tmp, "; " );

    if( (e = cookie_encode(l->str)) ) {
      tmp = strsav( tmp, e );
      freez( e );
    }
  }

  if( tmp ) {
    char *ret;

    ret = strdup( "Cookie: " );
    ret = strsav( ret, tmp );
    errlogf( HDR, "addh: %s\r\n", ret );
    enlist( csp->headers, ret );

    freez( tmp );
    freez( ret );
  }

} /* client_cookie_adder */

static void
client_xtra_adder( struct client_state *csp ) {

  struct list *l;

  for( l = xtra_list->next; l ; l = l->next ) {
    errlogf( HDR, "addh: %s\r\n", l->str );
    enlist( csp->headers, l->str );
  }

} /* client_xtra_adder */

static void
client_x_forwarded_adder(struct client_state *csp) {

  char *p = NULL;

  if( !add_forwarded )
    return;

  if( csp->x_forwarded ) {
    p = strsav( p, csp->x_forwarded );
    p = strsav( p, ", " );
    p = strsav( p, csp->ip_addr_str );
  } else {
    p = strsav( p, "X-Forwarded-For: " );
    p = strsav( p, csp->ip_addr_str );
  }

  errlogf( HDR, "addh: %s\r\n", p );

  enlist( csp->headers, p );

} /* client_x_forwarded_adder */
