/* This file is a component of the JunkBusterMT Proxy, based on original code from
 * the Internet Junkbuster Proxy.
 *
 * Copyright (C) 2000 Gabriel L. Somlo
 * Copyright (C) 1997 Anonymous Coders and Junkbusters Corporation
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see ../GPL.txt); if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/********************************  smrtcook.c  *********************************

    Purpose: Maintains the list of SmartCookies (cookies that are shared between
             all users of the proxy (for non-personal transactions where cookies
             are used by the server for unjustified tracking.

    Notes:   Whether SmartCookies apply to a www site is determined in the
             scookie.ini file, and the whole thing is handled elsewhere
             (loaders.c, parsers.c, etc.). Here we only maintain the cookie
             list, and rwlocked access to it via a strict interface.
**/

/*includes*/

#include <string.h>
#include <pthread.h>

#include "errlog.h"
#include "miscstr.h"


/*private defines (the public ones are in smrtcook.h)*/

/* -none- */


/*private macros (the public ones are in smrtcook.h)*/

/* -none- */


/*private typedefs (the public ones are in smrtcook.h)*/

/* the cookie-list type: */
struct cookie_list_t {
  char *cookie_text;		/* The NAME=VALUE field sent back to servers */
  char *expires;
  char *path;
  char *domain;
  int secure;
  struct cookie_list_t *next;
};


/*public global variables (to be declared "extern" in smrtcook.h)*/

/* -none- */


/*private global variables (only for this file, declare as "static")*/

static pthread_rwlock_t cookie_lock;	/* lock to protect the cookie list */

struct cookie_list_t smartcookie_list[1] = {	/* the cookie list */
  {NULL, NULL, NULL, NULL, 0, NULL}
};


/*private functions (only for this file, declare as "static")*/

/* -none- */


/*public functions (to be declared "extern" in smrtcook.h)*/

/*----------------------------------------------------------------------------*/
/* Initialize the SmartCookie module;                                         */
/* returns: void;                                                             */
/* parameters: none;                                                          */
/* notes: all this function does (for now) is to initialize the rwlock that   */
/*        will protect the cookie list; Therefore, it must be called by the   */
/*        main program before the multithreaded operation starts...           */
/*----------------------------------------------------------------------------*/
void
SmartCookieInit( void ) {

  pthread_rwlock_init( &cookie_lock, NULL );

} /* SmartCookieInit */


/*----------------------------------------------------------------------------*/
/* Insert/Replace a SmartCookie into the list;                                */
/* returns:    void;                                                          */
/* parameters: char *def_dom - in - default domain of the server setting the  */
/*                                  cookie;                                   */
/*             char *def_path - in - default path on server                   */
/*             char *cookiestr - in - the cookie string received from the     */
/*                                    server;                                 */
/* notes: the parameters are trated as read-only (we assume they will be      */
/*        freed after this call is completed. Only strdup's of (portions)     */
/*        of these strings will ever be used...                               */
/*----------------------------------------------------------------------------*/
void
SmartCookieInsertReplace( char *def_dom, char *def_path, char *cookiestr ) {

  struct cookie_list_t *cl, *c_new, *c_temp;
  char *cv[15], *cvv[3];
  int n, i, pathlen, namelen, replaced;

  /* Create new SmartCookie */
  if( !(c_new = zalloc(sizeof(struct cookie_list_t))) ) {
    errlogf( ERN, "SmartCookieInsertReplace(): malloc failed!\n" );
    exit( 1 );
  }

  /* parse the cookie-string */
  n = ssplit( cookiestr, "; ", cv, SZ(cv), 1, 1 );   /* get components */
  c_new->cookie_text = strdup( cv[0] );			/* main cookie text */
  for( i = 1; i < n; ++i ) {
    if( strncasecmp(cv[i], "expires=", 8) == 0 )
      c_new->expires = strdup( cv[i] + 8 );		/* expiration date */
    if( strncasecmp(cv[i], "path=",    5) == 0 )
      c_new->path    = strdup( cv[i] + 5 );		/* server path */
    if( strncasecmp(cv[i], "domain=",  7) == 0 )
      c_new->domain  = strdup( cv[i] + 7 );		/* server domain */
    if( strncasecmp(cv[i], "secure",   6) == 0 )
      c_new->secure  = 1;				/* cookie is secure */
  }
  if( !(c_new->domain) )
    c_new->domain = strdup( def_dom );	/* use default domain if not set */
  if( !(c_new->path) )
    c_new->path   = strdup( def_path );	/* use default path if not set */

  /* we'll need the length of the path in the process... */
  pathlen = strlen( c_new->path );

  /* we also need the length of the name, so we'll split up the main text */
  ssplit( cv[0], "=", cvv, SZ(cvv), 1, 1 );	/* need for err. check ?? */
  namelen = strlen( cvv[0] );

  /* we start messing with the global cookie list - write-lock it */
  pthread_rwlock_wrlock( &cookie_lock );

  /* Search for the proper place to insert into cookie list.
   * The cookie list is sorted in decreasing order of the path length.
   * We'll insert the item before all other items with equal path length.
   */

  /* insert new cookie into list */
  for( cl = smartcookie_list;
       (cl->next) && (pathlen < strlen(cl->next->path));
       cl = cl->next );	/* end for */
  c_new->next = cl->next;
  cl->next = c_new;

  replaced = 0;

  /* search and remove old cookie with same name/path/domain */
  for( cl = c_new;
       (cl->next) && (pathlen == strlen(cl->next->path));
       cl = cl->next
     ) {
    if( (strncmp(cvv[0], cl->next->cookie_text, namelen) == 0) &&
        (strcmp(c_new->path, cl->next->path) == 0) &&
        (strcmp(c_new->domain, cl->next->domain) == 0) ) {

      replaced = 1;

      c_temp = cl->next;
      cl->next = c_temp->next;
      freez( c_temp->cookie_text );
      freez( c_temp->expires );
      freez( c_temp->path );
      freez( c_temp->domain );
      freez( c_temp );

      /* This is REALLY, REALLY important!!! Leave out this "break" statement,
       * and we get a segfault !!! We only have one previous instance of the
       * cookie, so bailing out here is acceptable. Continuing with the for at
       * this point may result in a segfault because we've been mucking around
       * with cl and possibly pushed it outside the validity of the condition
       * "(cl->next != NULL )"
       */
      break;
    } /* if( ... ) */
  } /* for( cl ... ) */

  /* we're done messing with the global cookie list - unlock it */
  pthread_rwlock_unlock( &cookie_lock );

  /* log operation */
  errlogf( HDR, "\nSmartCookie: %s %s cookie list...",
                cvv[0], (replaced) ? "replaced in" : "added to" );

} /* SmartCookieInsertReplace */


/*----------------------------------------------------------------------------*/
/* Retrieve cookies matching the current domain and path;                     */
/* returns: char * - the string containing cookies separated by semicolons    */
/* parameters: char *cookiestr - in/out - the original string of semicolon    */
/*                                        separated cookies to add to         */
/*             char *dom - in - the domain of the server we're dealing with   */
/*             char *path - in - the path on the server we're dealing with    */
/* notes: the parameters dom and path are treated as read-only; the parameter */
/*        cookiestr will be potentially changed and returned as the result of */
/*        this function...                                                    */
/*----------------------------------------------------------------------------*/
char *
SmartCookieRetrieve( char *cookiestr, char *dom, char *path ) {

  int http_dlen, http_plen, cookie_dlen, cookie_plen;
  struct cookie_list_t *cl;

  /* we need the length of the server domain and path strings */
  http_dlen = strlen( dom );
  http_plen = strlen( path );

  /* we want to read cookies from the list - read-lock it */
  pthread_rwlock_rdlock( &cookie_lock );

  /* search cookie list for applicable cookies */
  for( cl = smartcookie_list->next; cl; cl = cl->next ) {

    /* we'll need the length of the cookie domain and path strings */
    cookie_dlen = strlen( cl->domain );
    cookie_plen = strlen( cl->path );

    /* compare domains and paths to see if the cookie is applicable */
    if( /* tail match of domains */
        (cookie_dlen <= http_dlen) &&
        (!strcasecmp(dom + (http_dlen - cookie_dlen), cl->domain)) &&
        /* head match of paths */
        (cookie_plen <= http_plen) &&
        (!strncmp(path, cl->path, cookie_plen)) ) {

      if( cookiestr )
        cookiestr = strsav( cookiestr, "; " );
      cookiestr = strsav( cookiestr, cl->cookie_text );

    } /* if( ...match... ) */

  } /* for( cl ... ) */

  /* we're done reading the cookie list - unlock it */
  pthread_rwlock_unlock( &cookie_lock );

  /* return the updated cookie string */
  return( cookiestr );

} /* SmartCookieRetrieve */
