/* This file is a component of the JunkBusterMT Proxy, based on original code from
 * the Internet Junkbuster Proxy.
 *
 * Copyright (C) 2000 Gabriel L. Somlo
 * Copyright (C) 1997 Anonymous Coders and Junkbusters Corporation
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see ../GPL.txt); if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>

#include "errlog.h"


static char *url_code_map[256];
static char *html_code_map[256];
static char *cookie_code_map[256];


/* the following two local functions will convert strings of (two) hex digits
 * into a decimal number. They are used in url_decode()
 */
static int
xdtoi( char d ) {

  if( (d >= '0') && (d <= '9') )
    return( d - '0' );
  if( (d >= 'a') && (d <= 'f') )
    return( d - 'a' + 10 );
  if( (d >= 'A') && (d <= 'F') )
    return( d - 'A' + 10 );

  return( 0 );

} /* xdtoi */

static int
xtoi( char *s ) {

  char d1, d2;
  int ret = 0;

  if( isxdigit((int)*s) ) {
    d1 = *s++;
    if( isxdigit((int)*s) ) {
      d2 = *s++;
      ret = 16 * xdtoi( d1 ) + xdtoi( d2 );
    }
  }

  return( ret );

} /* xtoi */


/* map_encode() will use the given map to encode the string (as requested by
 * url_encode(), cookie_encode(), and html_encode() ).
 */
static char *
map_encode( char **code_map, char *s ) {

  char *buf;
  unsigned char c, *p;
  char *m;

  /* Parameter sanity check */
  if( !s ) {
    errlogf( ERR, "map_encode(): called with NULL string!!!\n" );
    exit( 1 );	/* was return( NULL ); but I've decided not to be THAT nice...*/
  }

  /* each input char can expand to at most 6 chars */
  buf = malloc( (strlen(s) + 1) * 6 );

  for( p = buf; (c = *s); s++ )
    if( (m = code_map[c]) ) {
      strcpy( p, m );
      p += strlen( m );
    } else {
      *p++ = c;
    }

  *p = '\0';

  return( buf );

} /* map_encode */


/****************************** Public Functions ******************************/


/* Initialize code maps for url, cookie, and html encoding. This function must
 * be called from main() _before_ the threaded operation begins, to avoid
 * problems with reentrancy/thread safety. All the other functions ( *_encode(),
 * and url_decode() ) only read from the code maps, and thus cause no conflicts.
 */
void
init_code_maps( void ) {

  char tmp[BUFSIZ];
  int i;

  /* for cookies, ' ' -> '+' and turn commas and semicolons into hex */

  cookie_code_map[' '] = "+";

  sprintf( tmp, "%%%02X", ',' );
  cookie_code_map[','] = strdup( tmp );

  sprintf( tmp, "%%%02X", ';' );
  cookie_code_map[';'] = strdup( tmp );

  /* for url's, we do full URL encoding. */

  url_code_map[' '] = "+";	/* white-space gets turned into '+' */

  /* these punctuation chars map to themselves */
  url_code_map['-'] = "-";
  url_code_map['_'] = "_";
  url_code_map['.'] = ".";
  url_code_map['*'] = "*";
  url_code_map['@'] = "@";

  /* all other non-alphanumerics get turned into hex */
  for( i=0; i < 256; i++ )
    if( !isalnum(i) && !url_code_map[i] ) {
      sprintf( tmp, "%%%02X", i );
      url_code_map[i] = strdup( tmp );
    }

  /* for html, we encode the four "special" characters */
  html_code_map['"'] = "&quot;" ;
  html_code_map['&'] = "&amp;"  ;
  html_code_map['>'] = "&gt;"   ;
  html_code_map['<'] = "&lt;"   ;

} /* init_code_maps */


char *
url_encode( char *s ) {

  return( map_encode(url_code_map, s) );

} /* url_encode */


char *
cookie_encode( char *s ) {

  return( map_encode(cookie_code_map, s) );

} /* cookie_encode */


char *
html_encode( char *s ) {

  return( map_encode(html_code_map, s) );

} /* html_encode */


char *
url_decode( char *s ) {

  char *ret, *p, *q;

  /* Parameter sanity check */
  if( !s ) {
    errlogf( ERR, "url_decode(): called with NULL string!!!\n" );
    exit( 1 );	/* was return( NULL ); but I've decided not to be THAT nice...*/
  }

  ret = strdup( s );

  p = s;
  q = ret;

  while( *p )
    switch( *p ) {
      case '+':
        p++;
        *q++ = ' ';
        break;
      case '%':
        if( (*q = xtoi(p+1)) ) {
          p += 3;
          q++;
        } else {
          /* malformed, just use it */
          *q++ = *p++;
        }
        break;
      default:
        *q++ = *p++;
        break;
    }

  *q = '\0';
  return( ret );

} /* url_decode */
